@@ -2,28 +2,102 @@
 #include "sqlitedb.h"
 #include "sqlite.h"
 #include "Settings.h"
+#include "Data.h"
 
-#include <QDebug>
 #include <QMessageBox>
 #include <QApplication>
 #include <QTextCodec>
 #include <QMimeData>
 #include <QFile>
 #include <QUrl>
+#include <QtConcurrent/QtConcurrentRun>
+#include <QProgressDialog>
+
+#include "RowLoader.h"
 
 SqliteTableModel::SqliteTableModel(DBBrowserDB& db, QObject* parent, size_t chunkSize, const QString& encoding)
     : QAbstractTableModel(parent)
     , m_db(db)
-    , m_rowCount(0)
+    , m_lifeCounter(0)
+    , m_currentRowCount(0)
     , m_chunkSize(chunkSize)
-    , m_valid(false)
     , m_encoding(encoding)
 {
+    worker = new RowLoader(
+        [this](){ return m_db.get(tr("reading rows")); },
+        [this](QString stmt){ return m_db.logSQL(stmt, kLogMsg_App); },
+        m_headers, m_mutexDataCache, m_cache
+        );
+
+    worker->start();
+
+    // any UI updates must be performed in the UI thread, not in the worker thread:
+    connect(worker, &RowLoader::fetched, this, &SqliteTableModel::handleFinishedFetch, Qt::QueuedConnection);
+    connect(worker, &RowLoader::rowCountComplete, this, &SqliteTableModel::handleRowCountComplete, Qt::QueuedConnection);
+
     reset();
 }
 
+SqliteTableModel::~SqliteTableModel()
+{
+    worker->stop();
+    worker->wait();
+    worker->disconnect();
+    delete worker;
+}
+
+SqliteTableModel::RowCount SqliteTableModel::rowCountAvailable () const
+{
+    return m_rowCountAvailable;
+}
+
+void SqliteTableModel::handleFinishedFetch (int life_id, unsigned int fetched_row_begin, unsigned int fetched_row_end)
+{
+    if(life_id < m_lifeCounter)
+        return;
+
+    Q_ASSERT(fetched_row_end >= fetched_row_begin);
+
+    auto old_row_count = m_currentRowCount;
+
+    auto new_row_count = std::max(old_row_count, fetched_row_begin);
+    new_row_count = std::max(new_row_count, fetched_row_end);
+    Q_ASSERT(new_row_count >= old_row_count);
+
+    if(new_row_count != old_row_count)
+    {
+        beginInsertRows(QModelIndex(), old_row_count, new_row_count - 1);
+        m_currentRowCount = new_row_count;
+        endInsertRows();
+    }
+
+    if(fetched_row_end != fetched_row_begin)
+    {
+        // TODO optimize
+        int num_columns = m_headers.size();
+        emit dataChanged(createIndex(fetched_row_begin, 0), createIndex(fetched_row_end - 1, num_columns - 1));
+    }
+
+    if(m_rowCountAvailable != RowCount::Complete)
+        m_rowCountAvailable = RowCount::Partial;
+
+    emit finishedFetch(fetched_row_begin, fetched_row_end);
+}
+
+void SqliteTableModel::handleRowCountComplete (int life_id, int num_rows)
+{
+    if(life_id < m_lifeCounter)
+        return;
+
+    m_rowCountAvailable = RowCount::Complete;
+    handleFinishedFetch(life_id, num_rows, num_rows);
+}
+
 void SqliteTableModel::reset()
 {
+    beginResetModel();
+    clearCache();
+
     m_sTable.clear();
     m_sRowidColumn.clear();
     m_iSortColumn = 0;
@@ -33,6 +107,8 @@ void SqliteTableModel::reset()
     m_vDataTypes.clear();
     m_vDisplayFormat.clear();
     m_pseudoPk.clear();
+
+    endResetModel();
 }
 
 void SqliteTableModel::setChunkSize(size_t chunksize)
@@ -40,7 +116,7 @@ void SqliteTableModel::setChunkSize(size_t chunksize)
     m_chunkSize = chunksize;
 }
 
-void SqliteTableModel::setTable(const QString& table, int sortColumn, Qt::SortOrder sortOrder, const QVector<QString>& display_format)
+void SqliteTableModel::setTable(const sqlb::ObjectIdentifier& table, int sortColumn, Qt::SortOrder sortOrder, const QMap<int, QString> filterValues, const QVector<QString>& display_format)
 {
     // Unset all previous settings. When setting a table all information on the previously browsed data set is removed first.
     reset();
@@ -49,15 +125,18 @@ void SqliteTableModel::setTable(const QString& table, int sortColumn, Qt::SortOr
     m_sTable = table;
     m_vDisplayFormat = display_format;
 
+    for(auto filterIt=filterValues.constBegin(); filterIt!=filterValues.constEnd(); ++filterIt)
+        updateFilter(filterIt.key(), filterIt.value(), false);
+
     // The first column is the rowid column and therefore is always of type integer
     m_vDataTypes.push_back(SQLITE_INTEGER);
 
     // Get the data types of all other columns as well as the column names
     bool allOk = false;
-    if(m_db.getObjectByName(table)->type() == sqlb::Object::Types::Table)
+    if(m_db.getObjectByName(table) && m_db.getObjectByName(table)->type() == sqlb::Object::Types::Table)
     {
-        sqlb::TablePtr t = m_db.getObjectByName(table).dynamicCast<sqlb::Table>();
-        if(t && t->fields().size()) // parsing was OK
+        sqlb::TablePtr t = m_db.getObjectByName<sqlb::Table>(table);
+        if(t && t->fields.size()) // parsing was OK
         {
             m_sRowidColumn = t->rowidColumn();
             m_headers.push_back(m_sRowidColumn);
@@ -69,9 +148,9 @@ void SqliteTableModel::setTable(const QString& table, int sortColumn, Qt::SortOr
                     << "REAL"
                     << "TEXT"
                     << "BLOB";
-            foreach(const sqlb::FieldPtr fld,  t->fields())
+            for(const sqlb::Field& fld :  t->fields)
             {
-                QString name(fld->type().toUpper());
+                QString name(fld.type().toUpper());
                 int colType = dataTypes.indexOf(name);
                 colType = (colType == -1) ? SQLITE_TEXT : colType + 1;
                 m_vDataTypes.push_back(colType);
@@ -85,10 +164,10 @@ void SqliteTableModel::setTable(const QString& table, int sortColumn, Qt::SortOr
     // NOTE: It would be nice to eventually get rid of this piece here. As soon as the grammar parser is good enough...
     if(!allOk)
     {
-        QString sColumnQuery = QString::fromUtf8("SELECT * FROM %1;").arg(sqlb::escapeIdentifier(table));
+        QString sColumnQuery = QString::fromUtf8("SELECT * FROM %1;").arg(table.toString());
         m_sRowidColumn = "rowid";
         m_headers.push_back("rowid");
-        m_headers.append(getColumns(sColumnQuery, m_vDataTypes));
+        m_headers.append(getColumns(nullptr, sColumnQuery, m_vDataTypes));
     }
 
     // Set sort parameters. We're setting the sort column to an invalid value before calling sort() because this way, in sort() the
@@ -98,106 +177,35 @@ void SqliteTableModel::setTable(const QString& table, int sortColumn, Qt::SortOr
     sort(sortColumn, sortOrder);
 }
 
-namespace {
-QString rtrimChar(const QString& s, QChar c) {
-    QString r = s.trimmed();
-    while(r.endsWith(c))
-        r.chop(1);
-    return r;
-}
-}
-
 void SqliteTableModel::setQuery(const QString& sQuery, bool dontClearHeaders)
 {
     // clear
     if(!dontClearHeaders)
         reset();
+    else
+        clearCache();
 
     if(!m_db.isOpen())
         return;
 
     m_sQuery = sQuery.trimmed();
-
     removeCommentsFromQuery(m_sQuery);
 
-    // do a count query to get the full row count in a fast manner
-    m_rowCount = getQueryRowCount();
-    if(m_rowCount == -1)
-    {
-        m_valid = false;
-        return;
-    }
+    worker->setQuery(m_sQuery);
+    worker->triggerRowCountDetermination(m_lifeCounter);
 
-    // headers
     if(!dontClearHeaders)
-    {
-        m_headers.append(getColumns(sQuery, m_vDataTypes));
-    }
+        m_headers.append(getColumns(worker->getDb(), sQuery, m_vDataTypes));
 
     // now fetch the first entries
-    clearCache();
-    fetchData(0, m_chunkSize);
-    m_valid = true;
+    triggerCacheLoad(m_chunkSize / 2 - 1);
 
     emit layoutChanged();
 }
 
-int SqliteTableModel::getQueryRowCount()
-{
-    // Return -1 if there is an error
-    int retval = -1;
-
-    // Use a different approach of determining the row count when a EXPLAIN or a PRAGMA statement is used because a COUNT fails on these queries
-    if(m_sQuery.startsWith("EXPLAIN", Qt::CaseInsensitive) || m_sQuery.startsWith("PRAGMA", Qt::CaseInsensitive))
-    {
-        // So just execute the statement as it is and fetch all results counting the rows
-        sqlite3_stmt* stmt;
-        QByteArray utf8Query = m_sQuery.toUtf8();
-        if(sqlite3_prepare_v2(m_db._db, utf8Query, utf8Query.size(), &stmt, NULL) == SQLITE_OK)
-        {
-            retval = 0;
-            while(sqlite3_step(stmt) == SQLITE_ROW)
-                retval++;
-            sqlite3_finalize(stmt);
-
-            // Return the results but also set the chunk size the number of rows to prevent the lazy population mechanism to kick in as using LIMIT
-            // fails on this kind of queries as well
-            m_chunkSize = retval;
-            return retval;
-        }
-    } else {
-        // If it is a normal query - hopefully starting with SELECT - just do a COUNT on it and return the results
-        QString sCountQuery = QString("SELECT COUNT(*) FROM (%1);").arg(rtrimChar(m_sQuery, ';'));
-        m_db.logSQL(sCountQuery, kLogMsg_App);
-        QByteArray utf8Query = sCountQuery.toUtf8();
-
-        sqlite3_stmt* stmt;
-        int status = sqlite3_prepare_v2(m_db._db, utf8Query, utf8Query.size(), &stmt, NULL);
-        if(status == SQLITE_OK)
-        {
-            status = sqlite3_step(stmt);
-            if(status == SQLITE_ROW)
-            {
-                QString sCount = QString::fromUtf8((const char*)sqlite3_column_text(stmt, 0));
-                retval = sCount.toInt();
-            }
-            sqlite3_finalize(stmt);
-        } else {
-            qWarning() << "Count query failed: " << sCountQuery;
-        }
-    }
-
-    return retval;
-}
-
 int SqliteTableModel::rowCount(const QModelIndex&) const
 {
-    return m_data.size(); // current fetched row count
-}
-
-int SqliteTableModel::totalRowCount() const
-{
-    return m_rowCount;
+    return m_currentRowCount;
 }
 
 int SqliteTableModel::columnCount(const QModelIndex&) const
@@ -205,6 +213,11 @@ int SqliteTableModel::columnCount(const QModelIndex&) const
     return m_headers.size();
 }
 
+int SqliteTableModel::filterCount() const
+{
+    return m_mWhere.size();
+}
+
 QVariant SqliteTableModel::headerData(int section, Qt::Orientation orientation, int role) const
 {
     if (role != Qt::DisplayRole)
@@ -227,23 +240,37 @@ QVariant SqliteTableModel::data(const QModelIndex &index, int role) const
     if (!index.isValid())
         return QVariant();
 
-    if (index.row() >= m_rowCount)
+    if (index.row() >= rowCount())
         return QVariant();
 
+    QMutexLocker lock(&m_mutexDataCache);
+
+    Row blank_data;
+    bool row_available;
+
+    const Row * cached_row;
+    if(m_cache.count(index.row()))
+    {
+        cached_row = &m_cache.at(index.row());
+        row_available = true;
+    } else {
+        blank_data = makeDefaultCacheEntry();
+        cached_row = &blank_data;
+        row_available = false;
+    }
+
     if(role == Qt::DisplayRole || role == Qt::EditRole)
     {
-        // If this row is not in the cache yet get it first
-        while(index.row() >= m_data.size() && canFetchMore())
-            const_cast<SqliteTableModel*>(this)->fetchMore();   // Nothing evil to see here, move along
-
-        if(role == Qt::DisplayRole && m_data.at(index.row()).at(index.column()).isNull())
+        if(!row_available)
+            return tr("loading...");
+        if(role == Qt::DisplayRole && cached_row->at(index.column()).isNull())
         {
             return Settings::getValue("databrowser", "null_text").toString();
-        } else if(role == Qt::DisplayRole && isBinary(index)) {
-            return "BLOB";
+        } else if(role == Qt::DisplayRole && nosync_isBinary(index)) {
+            return Settings::getValue("databrowser", "blob_text").toString();
         } else if(role == Qt::DisplayRole) {
             int limit = Settings::getValue("databrowser", "symbol_limit").toInt();
-            QByteArray displayText = m_data.at(index.row()).at(index.column());
+            QByteArray displayText = cached_row->at(index.column());
             if (displayText.length() > limit) {
                 // Add "..." to the end of truncated strings
                 return decode(displayText.left(limit).append(" ..."));
@@ -251,34 +278,39 @@ QVariant SqliteTableModel::data(const QModelIndex &index, int role) const
                 return decode(displayText);
             }
         } else {
-            return decode(m_data.at(index.row()).at(index.column()));
+            return decode(cached_row->at(index.column()));
         }
     } else if(role == Qt::FontRole) {
         QFont font;
-        if(m_data.at(index.row()).at(index.column()).isNull() || isBinary(index))
+        if(!row_available || cached_row->at(index.column()).isNull() || nosync_isBinary(index))
             font.setItalic(true);
         return font;
     } else if(role == Qt::ForegroundRole) {
-        if(m_data.at(index.row()).at(index.column()).isNull())
+        if(!row_available)
+            return QColor(100, 100, 100);
+        if(cached_row->at(index.column()).isNull())
             return QColor(Settings::getValue("databrowser", "null_fg_colour").toString());
-        else if (isBinary(index))
+        else if (nosync_isBinary(index))
             return QColor(Settings::getValue("databrowser", "bin_fg_colour").toString());
         return QColor(Settings::getValue("databrowser", "reg_fg_colour").toString());
     } else if (role == Qt::BackgroundRole) {
-        if(m_data.at(index.row()).at(index.column()).isNull())
+        if(!row_available)
+            return QColor(255, 200, 200);
+        if(cached_row->at(index.column()).isNull())
             return QColor(Settings::getValue("databrowser", "null_bg_colour").toString());
-        else if (isBinary(index))
+        else if (nosync_isBinary(index))
             return QColor(Settings::getValue("databrowser", "bin_bg_colour").toString());
         return QColor(Settings::getValue("databrowser", "reg_bg_colour").toString());
     } else if(role == Qt::ToolTipRole) {
         sqlb::ForeignKeyClause fk = getForeignKeyClause(index.column()-1);
         if(fk.isSet())
-            return tr("References %1(%2)\nHold Ctrl+Shift and click to jump there").arg(fk.table()).arg(fk.columns().join(","));
+          return tr("References %1(%2)\nHold %3Shift and click to jump there").arg(fk.table()).arg(fk.columns().join(","))
+            .arg(QKeySequence(Qt::CTRL).toString(QKeySequence::NativeText));
         else
             return QString();
-    } else {
-        return QVariant();
     }
+
+    return QVariant();
 }
 
 sqlb::ForeignKeyClause SqliteTableModel::getForeignKeyClause(int column) const
@@ -298,15 +330,15 @@ sqlb::ForeignKeyClause SqliteTableModel::getForeignKeyClause(int column) const
         return empty_foreign_key_clause;
 
     // Convert object to a table and check if the column number is in the valid range
-    sqlb::TablePtr tbl = obj.dynamicCast<sqlb::Table>();
-    if(tbl && tbl->name().size() && (column >= 0 && column < tbl->fields().count()))
+    sqlb::TablePtr tbl = std::dynamic_pointer_cast<sqlb::Table>(obj);
+    if(tbl && tbl->name().size() && (column >= 0 && column < static_cast<int>(tbl->fields.size())))
     {
         // Note that the rowid column has number -1 here, it can safely be excluded since there will never be a
         // foreign key on that column.
 
-        sqlb::ConstraintPtr ptr = tbl->constraint({tbl->fields().at(column)}, sqlb::Constraint::ForeignKeyConstraintType);
+        sqlb::ConstraintPtr ptr = tbl->constraint({tbl->fields.at(column).name()}, sqlb::Constraint::ForeignKeyConstraintType);
         if(ptr)
-            return *(ptr.dynamicCast<sqlb::ForeignKeyClause>());
+            return *(std::dynamic_pointer_cast<sqlb::ForeignKeyClause>(ptr));
     }
 
     return empty_foreign_key_clause;
@@ -325,20 +357,29 @@ bool SqliteTableModel::setData(const QModelIndex& index, const QVariant& value,
 
 bool SqliteTableModel::setTypedData(const QModelIndex& index, bool isBlob, const QVariant& value, int role)
 {
+    if(readingData()) {
+        // can't insert rows while reading data in background
+        return false;
+    }
+
     if(index.isValid() && role == Qt::EditRole)
     {
+        QMutexLocker lock(&m_mutexDataCache);
+
+        auto & cached_row = m_cache.at(index.row());
+
         QByteArray newValue = encode(value.toByteArray());
-        QByteArray oldValue = m_data.at(index.row()).at(index.column());
+        QByteArray oldValue = cached_row.at(index.column());
 
         // Special handling for integer columns: instead of setting an integer column to an empty string, set it to '0' when it is also
         // used in a primary key. Otherwise SQLite will always output an 'datatype mismatch' error.
         if(newValue == "" && !newValue.isNull())
         {
-            sqlb::TablePtr table = m_db.getObjectByName(m_sTable).dynamicCast<sqlb::Table>();
+            sqlb::TablePtr table = m_db.getObjectByName<sqlb::Table>(m_sTable);
             if(table)
             {
-                sqlb::FieldPtr field = table->field(table->findField(m_headers.at(index.column())));
-                if(table->primaryKey().contains(field) && field->isInteger())
+                auto field = sqlb::findField(table, m_headers.at(index.column()));
+                if(contains(table->primaryKey(), field->name()) && field->isInteger())
                     newValue = "0";
             }
         }
@@ -348,16 +389,22 @@ bool SqliteTableModel::setTypedData(const QModelIndex& index, bool isBlob, const
         if(oldValue == newValue && oldValue.isNull() == newValue.isNull())
             return true;
 
-        if(m_db.updateRecord(m_sTable, m_headers.at(index.column()), m_data[index.row()].at(0), newValue, isBlob, m_pseudoPk))
+        if(m_db.updateRecord(m_sTable, m_headers.at(index.column()), cached_row.at(0), newValue, isBlob, m_pseudoPk))
         {
-            // Only update the cache if this row has already been read, if not there's no need to do any changes to the cache
-            if(index.row() < m_data.size())
-                m_data[index.row()].replace(index.column(), newValue);
-
-            emit(dataChanged(index, index));
+            cached_row.replace(index.column(), newValue);
+            if(m_headers.at(index.column()) == m_sRowidColumn) {
+                cached_row.replace(0, newValue);
+                const QModelIndex& rowidIndex = index.sibling(index.row(), 0);
+                lock.unlock();
+                emit dataChanged(rowidIndex, rowidIndex);
+            } else {
+                lock.unlock();
+            }
+            emit dataChanged(index, index);
             return true;
         } else {
-            QMessageBox::warning(0, qApp->applicationName(), tr("Error changing data:\n%1").arg(m_db.lastError()));
+            lock.unlock();
+            QMessageBox::warning(nullptr, qApp->applicationName(), tr("Error changing data:\n%1").arg(m_db.lastError()));
             return false;
         }
     }
@@ -365,17 +412,6 @@ bool SqliteTableModel::setTypedData(const QModelIndex& index, bool isBlob, const
     return false;
 }
 
-bool SqliteTableModel::canFetchMore(const QModelIndex&) const
-{
-    return m_data.size() < m_rowCount;
-}
-
-void SqliteTableModel::fetchMore(const QModelIndex&)
-{
-    int row = m_data.size();
-    fetchData(row, row + m_chunkSize);
-}
-
 Qt::ItemFlags SqliteTableModel::flags(const QModelIndex& index) const
 {
     if(!index.isValid())
@@ -387,9 +423,8 @@ Qt::ItemFlags SqliteTableModel::flags(const QModelIndex& index) const
     bool custom_display_format = false;
     if(m_vDisplayFormat.size())
     {
-        // NOTE: This assumes that custom display formats never start and end with a backtick
         if(index.column() > 0)
-            custom_display_format = !(m_vDisplayFormat.at(index.column()-1).startsWith("`") && m_vDisplayFormat.at(index.column()-1).endsWith("`"));
+            custom_display_format = m_vDisplayFormat.at(index.column()-1) != sqlb::escapeIdentifier(headerData(index.column(), Qt::Horizontal).toString());
     }
 
     if(!isBinary(index) && !custom_display_format)
@@ -409,20 +444,38 @@ void SqliteTableModel::sort(int column, Qt::SortOrder order)
     m_sSortOrder = (order == Qt::AscendingOrder ? "ASC" : "DESC");
 
     // Set the new query (but only if a table has already been set
-    if(m_sTable != "")
+    if(!m_sTable.isEmpty())
         buildQuery();
 }
 
+SqliteTableModel::Row SqliteTableModel::makeDefaultCacheEntry () const
+{
+    Row blank_data;
+
+    for(int i=0; i < m_headers.size(); ++i)
+        blank_data.push_back("");
+
+    return blank_data;
+}
+
+bool SqliteTableModel::readingData() const
+{
+    return worker->readingData();
+}
+
 bool SqliteTableModel::insertRows(int row, int count, const QModelIndex& parent)
 {
     if(!isEditable())
         return false;
 
-    QByteArrayList blank_data;
-    for(int i=0; i < m_headers.size(); ++i)
-        blank_data.push_back("");
+    if(readingData()) {
+        // can't insert rows while reading data in background
+        return false;
+    }
 
-    DataType tempList;
+    const auto blank_data = makeDefaultCacheEntry();
+
+    std::vector<Row> tempList;
     for(int i=row; i < row + count; ++i)
     {
         QString rowid = m_db.addRecord(m_sTable);
@@ -430,27 +483,28 @@ bool SqliteTableModel::insertRows(int row, int count, const QModelIndex& parent)
         {
             return false;
         }
-        m_rowCount++;
-        tempList.append(blank_data);
-        tempList[i - row].replace(0, rowid.toUtf8());
+        tempList.push_back(blank_data);
+        tempList.back().replace(0, rowid.toUtf8());
 
         // update column with default values
-        QByteArrayList rowdata;
-        if( m_db.getRow(m_sTable, rowid, rowdata) )
+        Row rowdata;
+        if(m_db.getRow(m_sTable, rowid, rowdata))
         {
             for(int j=1; j < m_headers.size(); ++j)
             {
-                tempList[i - row].replace(j, rowdata[j - 1]);
+                tempList.back().replace(j, rowdata[j - 1]);
             }
         }
     }
 
     beginInsertRows(parent, row, row + count - 1);
-    for(int i = 0; i < tempList.size(); ++i)
+    for(unsigned int i = 0; i < tempList.size(); ++i)
     {
-        m_data.insert(i + row, tempList.at(i));
+        m_cache.insert(i + row, std::move(tempList.at(i)));
+        m_currentRowCount++;
     }
     endInsertRows();
+
     return true;
 }
 
@@ -459,23 +513,32 @@ bool SqliteTableModel::removeRows(int row, int count, const QModelIndex& parent)
     if(!isEditable())
         return false;
 
-    beginRemoveRows(parent, row, row + count - 1);
-
-    bool ok = true;
+    if(readingData()) {
+        // can't delete rows while reading data in background
+        return false;
+    }
 
     QStringList rowids;
     for(int i=count-1;i>=0;i--)
     {
-        rowids.append(m_data.at(row + i).at(0));
-        m_data.removeAt(row + i);
-        --m_rowCount;
-    }
-    if(!m_db.deleteRecords(m_sTable, rowids))
-    {
-        ok = false;
+        if(m_cache.count(row+i)) {
+            rowids.append(m_cache.at(row + i).at(0));
+        }
     }
 
-    endRemoveRows();
+    bool ok = m_db.deleteRecords(m_sTable, rowids, m_pseudoPk);
+
+    if (ok) {
+        beginRemoveRows(parent, row, row + count - 1);
+
+        for(int i=count-1;i>=0;i--)
+        {
+            m_cache.erase(row + i);
+            m_currentRowCount--;
+        }
+
+        endRemoveRows();
+    }
     return ok;
 }
 
@@ -484,15 +547,17 @@ QModelIndex SqliteTableModel::dittoRecord(int old_row)
     if(!isEditable())
         return QModelIndex();
 
-    insertRow(rowCount());
+    if (!insertRow(rowCount()))
+        return QModelIndex();
+
     int firstEditedColumn = 0;
     int new_row = rowCount() - 1;
 
-    sqlb::TablePtr t = m_db.getObjectByName(m_sTable).dynamicCast<sqlb::Table>();
+    sqlb::TablePtr t = m_db.getObjectByName<sqlb::Table>(m_sTable);
 
-    sqlb::FieldVector pk = t->primaryKey();
-    for (int col = 0; col < t->fields().size(); ++col) {
-        if(!pk.contains(t->fields().at(col))) {
+    QStringList pk = t->primaryKey();
+    for (size_t col = 0; col < t->fields.size(); ++col) {
+        if(!contains(pk, t->fields.at(col).name())) {
             if (!firstEditedColumn)
                 firstEditedColumn = col + 1;
 
@@ -504,61 +569,7 @@ QModelIndex SqliteTableModel::dittoRecord(int old_row)
     return index(new_row, firstEditedColumn);
 }
 
-void SqliteTableModel::fetchData(unsigned int from, unsigned to)
-{
-    int currentsize = m_data.size();
-
-    QString sLimitQuery;
-    if(m_sQuery.startsWith("PRAGMA", Qt::CaseInsensitive) || m_sQuery.startsWith("EXPLAIN", Qt::CaseInsensitive))
-    {
-        sLimitQuery = m_sQuery;
-    } else {
-        // Remove trailing trailing semicolon
-        QString queryTemp = rtrimChar(m_sQuery, ';');
-
-        // If the query ends with a LIMIT statement take it as it is, if not append our own LIMIT part for lazy population
-        if(queryTemp.contains(QRegExp("LIMIT\\s+.+\\s*((,|\\b(OFFSET)\\b)\\s*.+\\s*)?$", Qt::CaseInsensitive)))
-            sLimitQuery = queryTemp;
-        else
-            sLimitQuery = queryTemp + QString(" LIMIT %1, %2;").arg(from).arg(to-from);
-    }
-    m_db.logSQL(sLimitQuery, kLogMsg_App);
-    QByteArray utf8Query = sLimitQuery.toUtf8();
-    sqlite3_stmt *stmt;
-    int status = sqlite3_prepare_v2(m_db._db, utf8Query, utf8Query.size(), &stmt, NULL);
-
-    if(SQLITE_OK == status)
-    {
-        while(sqlite3_step(stmt) == SQLITE_ROW)
-        {
-            QByteArrayList rowdata;
-            for (int i = 0; i < m_headers.size(); ++i)
-            {
-                if(sqlite3_column_type(stmt, i) == SQLITE_NULL)
-                {
-                    rowdata.append(QByteArray());
-                } else {
-                    int bytes = sqlite3_column_bytes(stmt, i);
-                    if(bytes)
-                        rowdata.append(QByteArray(static_cast<const char*>(sqlite3_column_blob(stmt, i)), bytes));
-                    else
-                        rowdata.append(QByteArray(""));
-                }
-            }
-            m_data.push_back(rowdata);
-        }
-    }
-    sqlite3_finalize(stmt);
-
-    // Check if there was any new data
-    if(m_data.size() > currentsize)
-    {
-        beginInsertRows(QModelIndex(), currentsize, m_data.size()-1);
-        endInsertRows();
-    }
-}
-
-void SqliteTableModel::buildQuery()
+QString SqliteTableModel::customQuery(bool withRowid)
 {
     QString where;
 
@@ -568,12 +579,10 @@ void SqliteTableModel::buildQuery()
 
         for(QMap<int, QString>::const_iterator i=m_mWhere.constBegin();i!=m_mWhere.constEnd();++i)
         {
-            QString column;
-            if(m_vDisplayFormat.size())
-                column = QString("col%1").arg(i.key());
-            else
-                column = m_headers.at(i.key());
-            where.append(QString("%1 %2 AND ").arg(sqlb::escapeIdentifier(column)).arg(i.value()));
+            QString columnId = sqlb::escapeIdentifier(m_headers.at(i.key()));
+            if(m_vDisplayFormat.size() && m_vDisplayFormat.at(i.key()-1) != columnId)
+                columnId = m_vDisplayFormat.at(i.key()-1);
+            where.append(QString("%1 %2 AND ").arg(columnId).arg(i.value()));
         }
 
         // Remove last 'AND '
@@ -581,12 +590,21 @@ void SqliteTableModel::buildQuery()
     }
 
     QString selector;
+    if (withRowid)
+        selector = sqlb::escapeIdentifier(m_headers.at(0)) + ",";
+
     if(m_vDisplayFormat.empty())
     {
-        selector = "*";
+        selector += "*";
     } else {
-        for(int i=0;i<m_vDisplayFormat.size();i++)
-            selector += m_vDisplayFormat.at(i) + " AS " + QString("col%1").arg(i+1) + ",";
+        QString columnId;
+        for(int i=0;i<m_vDisplayFormat.size();i++) {
+            columnId = sqlb::escapeIdentifier(m_headers.at(i+1));
+            if (columnId != m_vDisplayFormat.at(i))
+                selector += m_vDisplayFormat.at(i) + " AS " + columnId + ",";
+            else
+                selector += columnId + ",";
+        }
         selector.chop(1);
     }
 
@@ -594,18 +612,24 @@ void SqliteTableModel::buildQuery()
     // The reason is that we're adding '%' characters automatically around search terms (and even if we didn't the user could add
     // them manually) which means that e.g. searching for '1' results in another '%1' in the string which then totally confuses
     // the QString::arg() function, resulting in an invalid SQL.
-    QString sql = QString("SELECT %1,%2 FROM %3 ")
-            .arg(sqlb::escapeIdentifier(m_headers.at(0)))
+    return QString("SELECT %1 FROM %2 ")
             .arg(selector)
-            .arg(sqlb::escapeIdentifier(m_sTable))
+            .arg(m_sTable.toString())
             + where
             + QString("ORDER BY %1 %2")
             .arg(sqlb::escapeIdentifier(m_headers.at(m_iSortColumn)))
             .arg(m_sSortOrder);
-    setQuery(sql, true);
 }
 
-void SqliteTableModel::removeCommentsFromQuery(QString& query) {
+void SqliteTableModel::buildQuery()
+{
+    setQuery(customQuery(true), true);
+}
+
+void SqliteTableModel::removeCommentsFromQuery(QString& query)
+{
+    int oldSize = query.size();
+
     // first remove block comments
     {
         QRegExp rxSQL("^((?:(?:[^'/]|/(?![*]))*|'[^']*')*)(/[*](?:[^*]|[*](?!/))*[*]/)(.*)$");	// set up regex to find block comment
@@ -636,8 +660,8 @@ void SqliteTableModel::removeCommentsFromQuery(QString& query) {
          *   (?:                 |           )*         # code is none or many strings alternating with non-strings
          *                        (?:'[^']*')           # a string is a quote, followed by none or more non-quotes, followed by a quote
          *      (?:[^'-]|-(?!-))*                       # non-string is a sequence of characters which aren't quotes or hyphens,
-         *                                                OR if they are hyphens then they can't be followed immediately by another hyphen
          */
+
         QRegExp rxSQL("^((?:(?:[^'-]|-(?!-))*|(?:'[^']*'))*)(--[^\\r\\n]*)([\\r\\n]*)(.*)$");	// set up regex to find end-of-line comment
         QString result;
 
@@ -657,18 +681,23 @@ void SqliteTableModel::removeCommentsFromQuery(QString& query) {
         query = result.trimmed();
     }
 
-    // Remove multiple line breaks that might have been created by deleting comments till the end of the line but not including the line break
-    query.replace(QRegExp("\\n+"), "\n");
+    if (oldSize != query.size()) {
+        // Remove multiple line breaks that might have been created by deleting comments till the end of the line but not including the line break
+        query.replace(QRegExp("\\n+"), "\n");
 
-    // Also remove any remaining whitespace at the end of each line
-    query.replace(QRegExp("[ \t]+\n"), "\n");
+        // Also remove any remaining whitespace at the end of each line
+        query.replace(QRegExp("[ \t]+\n"), "\n");
+    }
 }
 
-QStringList SqliteTableModel::getColumns(const QString& sQuery, QVector<int>& fieldsTypes)
+QStringList SqliteTableModel::getColumns(std::shared_ptr<sqlite3> pDb, const QString& sQuery, QVector<int>& fieldsTypes)
 {
+    if(!pDb)
+        pDb = m_db.get(tr("retrieving list of columns"));
+
     sqlite3_stmt* stmt;
     QByteArray utf8Query = sQuery.toUtf8();
-    int status = sqlite3_prepare_v2(m_db._db, utf8Query, utf8Query.size(), &stmt, NULL);
+    int status = sqlite3_prepare_v2(pDb.get(), utf8Query, utf8Query.size(), &stmt, nullptr);
     QStringList listColumns;
     if(SQLITE_OK == status)
     {
@@ -685,13 +714,14 @@ QStringList SqliteTableModel::getColumns(const QString& sQuery, QVector<int>& fi
     return listColumns;
 }
 
-void SqliteTableModel::updateFilter(int column, const QString& value)
+void SqliteTableModel::updateFilter(int column, const QString& value, bool applyQuery)
 {
     // Check for any special comparison operators at the beginning of the value string. If there are none default to LIKE.
     QString op = "LIKE";
     QString val, val2;
     QString escape;
     bool numeric = false, ok = false;
+
     // range/BETWEEN operator
     if (value.contains("~")) {
         int sepIdx = value.indexOf('~');
@@ -714,30 +744,25 @@ void SqliteTableModel::updateFilter(int column, const QString& value)
             // Check if we're filtering for '<> NULL'. In this case we need a special comparison operator.
             if(value.left(2) == "<>" && value.mid(2) == "NULL")
             {
-                // We are filtering for '<> NULL'. Override the comparison operator to search for NULL values in this column. Also treat search value (NULL) as number,
+                // We are filtering for '<>NULL'. Override the comparison operator to search for NULL values in this column. Also treat search value (NULL) as number,
                 // in order to avoid putting quotes around it.
                 op = "IS NOT";
                 numeric = true;
                 val = "NULL";
+            } else if(value.left(2) == "<>" && value.mid(2) == "''") {
+                // We are filtering for "<>''", i.e. for everything which is not an empty string
+                op = "<>";
+                numeric = true;
+                val = "''";
             } else {
-                bool ok;
-                value.mid(2).toFloat(&ok);
-                if(ok)
-                {
-                    op = value.left(2);
-                    val = value.mid(2);
-                    numeric = true;
-                }
+                value.mid(2).toFloat(&numeric);
+                op = value.left(2);
+                val = value.mid(2);
             }
         } else if(value.left(1) == ">" || value.left(1) == "<") {
-            bool ok;
-            value.mid(1).toFloat(&ok);
-            if(ok)
-            {
-                op = value.left(1);
-                val = value.mid(1);
-                numeric = true;
-            }
+            value.mid(1).toFloat(&numeric);
+            op = value.left(1);
+            val = value.mid(1);
         } else if(value.left(1) == "=") {
             val = value.mid(1);
 
@@ -774,13 +799,15 @@ void SqliteTableModel::updateFilter(int column, const QString& value)
     }
     if(val.isEmpty())
         val = value;
-    if(!numeric)
-        val = QString("'%1'").arg(val.replace("'", "''"));
 
     // If the value was set to an empty string remove any filter for this column. Otherwise insert a new filter rule or replace the old one if there is already one
-    if(val == "''" || val == "'%'" || val == "'%%'")
+    if(val == "" || val == "%" || val == "%%")
         m_mWhere.remove(column);
     else {
+        // Quote and escape value, but only if it's not numeric and not the empty string sequence
+        if(!numeric && val != "''")
+            val = QString("'%1'").arg(val.replace("'", "''"));
+
         QString whereClause(op + " " + QString(encode(val.toUtf8())));
         if (!val2.isEmpty())
             whereClause += " AND " + QString(encode(val2.toUtf8()));
@@ -789,22 +816,44 @@ void SqliteTableModel::updateFilter(int column, const QString& value)
     }
 
     // Build the new query
-    buildQuery();
+    if (applyQuery)
+        buildQuery();
 }
 
 void SqliteTableModel::clearCache()
 {
-	if(!m_data.empty())
-	{
-		beginRemoveRows(QModelIndex(), 0, m_data.size() - 1);
-		m_data.clear();
-		endRemoveRows();
-	}
+    m_lifeCounter++;
+
+    if(m_db.isOpen()) {
+        worker->cancel();
+        worker->waitUntilIdle();
+    }
+
+    if(m_currentRowCount > 0)
+    {
+        beginRemoveRows(QModelIndex(), 0, m_currentRowCount - 1);
+        endRemoveRows();
+    }
+
+    m_cache.clear();
+    m_currentRowCount = 0;
+    m_rowCountAvailable = RowCount::Unknown;
 }
 
 bool SqliteTableModel::isBinary(const QModelIndex& index) const
 {
-    return m_data.at(index.row()).at(index.column()).left(1024).contains('\0');
+    QMutexLocker lock(&m_mutexDataCache);
+    return nosync_isBinary(index);
+}
+
+bool SqliteTableModel::nosync_isBinary(const QModelIndex& index) const
+{
+    if(!m_cache.count(index.row()))
+        return false;
+
+    const auto & cached_row = m_cache.at(index.row());
+
+    return !isTextOnly(cached_row.at(index.column()), m_encoding, true);
 }
 
 QByteArray SqliteTableModel::encode(const QByteArray& str) const
@@ -851,6 +900,10 @@ bool SqliteTableModel::dropMimeData(const QMimeData* data, Qt::DropAction, int r
 
 void SqliteTableModel::setPseudoPk(const QString& pseudoPk)
 {
+    // Do nothing if the value didn't change
+    if(m_pseudoPk == pseudoPk)
+        return;
+
     if(pseudoPk.isEmpty())
     {
         m_pseudoPk.clear();
@@ -867,5 +920,73 @@ void SqliteTableModel::setPseudoPk(const QString& pseudoPk)
 
 bool SqliteTableModel::isEditable() const
 {
-    return !m_sTable.isEmpty();
+    return !m_sTable.isEmpty() &&
+            m_db.isOpen() &&
+            ((m_db.getObjectByName(m_sTable) && m_db.getObjectByName(m_sTable)->type() == sqlb::Object::Types::Table) || !m_pseudoPk.isEmpty());
+}
+
+void SqliteTableModel::triggerCacheLoad (int row) const
+{
+    size_t row_begin = std::max(0, row - int(m_chunkSize) / 2);
+    size_t row_end = row + m_chunkSize / 2;
+
+    if(rowCountAvailable() == RowCount::Complete) {
+        row_end = std::min(row_end, size_t(rowCount()));
+    } else {
+        // will be truncated by reader
+    }
+
+    // avoid re-fetching data
+    QMutexLocker lk(&m_mutexDataCache);
+    m_cache.smallestNonAvailableRange(row_begin, row_end);
+
+    if(row_end != row_begin)
+        worker->triggerFetch(m_lifeCounter, row_begin, row_end);
+}
+
+void SqliteTableModel::triggerCacheLoad (int row_begin, int row_end) const
+{
+    if(row_end == row_begin)
+        return;
+
+    triggerCacheLoad((row_begin + row_end) / 2);
+}
+
+bool SqliteTableModel::completeCache () const
+{
+    // Show progress dialog because fetching all data might take some time but only show
+    // cancel button if we allow cancellation here. This isn't
+    QProgressDialog progress(tr("Fetching data..."),
+                             tr("Cancel"), 0, rowCount());
+    progress.setWindowModality(Qt::ApplicationModal);
+    progress.show();
+
+    waitUntilIdle();
+
+    // This loop fetches all data by loading it block by block into the cache
+    for(int i=0;i<rowCount()+static_cast<int>(m_chunkSize)/2;i+=m_chunkSize)
+    {
+        progress.setValue(i);
+        qApp->processEvents();
+        if(progress.wasCanceled())
+            return false;
+
+        triggerCacheLoad(i);
+        worker->waitUntilIdle();
+    }
+
+    return true;
+}
+
+bool SqliteTableModel::isCacheComplete () const
+{
+    if(readingData())
+        return false;
+    QMutexLocker lock(&m_mutexDataCache);
+    return m_cache.numSet() == m_currentRowCount;
+}
+
+void SqliteTableModel::waitUntilIdle () const
+{
+    worker->waitUntilIdle();
 }
